@@ -29,28 +29,38 @@ import shutil
 import os
 import pprint
 import configparser
+import json
+import base64
 from git.config import GitConfigParser
 from .libkas import (ssh_cleanup_agent, ssh_setup_agent, ssh_no_host_key_check,
                      get_build_environ, repos_fetch, repos_apply_patches)
+from .context import ManagedEnvironment as ME
+from .context import get_context
 from .includehandler import IncludeException
 from .kasusererror import ArgsCombinationError
 
 __license__ = 'MIT'
 __copyright__ = 'Copyright (c) Siemens AG, 2017-2018'
 
+KAS_USER_NAME = 'kas User'
+KAS_USER_EMAIL = 'kas@example.com'
+HTTPS_PORT_DEFAULT = 443
+SSH_PORT_DEFAULT = 22
+
 
 class Macro:
     """
         Contains commands and provides method to run them.
     """
 
-    def __init__(self, use_common_setup=True, use_common_cleanup=True):
+    def __init__(self, use_common_setup=True):
         if use_common_setup:
             repo_loop = Loop('repo_setup_loop')
             repo_loop.add(SetupReposStep())
 
+            # setup commands are pairs of setup / cleanup commands
             self.setup_commands = [
-                SetupDir(),
+                (SetupDir(), None)
             ]
 
             if ('SSH_PRIVATE_KEY' in os.environ
@@ -59,9 +69,10 @@ class Macro:
                     raise ArgsCombinationError(
                         'Internal SSH agent (e.g. for "SSH_PRIVATE_KEY") can '
                         'only be started if no external one is passed.')
-                self.setup_commands.append(SetupSSHAgent())
+                self.setup_commands.append((SetupSSHAgent(),
+                                            CleanupSSHAgent()))
 
-            self.setup_commands += [
+            self.setup_commands += [(x, None) for x in [
                 SetupHome(),
                 InitSetupRepos(),
                 repo_loop,
@@ -70,19 +81,10 @@ class Macro:
                 ReposApplyPatches(),
                 SetupEnviron(),
                 WriteBBConfig(),
-            ]
+            ]]
         else:
             self.setup_commands = []
 
-        if (use_common_cleanup
-                and ('SSH_PRIVATE_KEY' in os.environ
-                     or 'SSH_PRIVATE_KEY_FILE' in os.environ)):
-            self.cleanup_commands = [
-                CleanupSSHAgent(),
-            ]
-        else:
-            self.cleanup_commands = []
-
         self.commands = []
 
     def add(self, command):
@@ -96,15 +98,24 @@ class Macro:
             Runs a command from the command list with respect to the
             configuration.
         """
-        skip = skip or []
-        joined_commands = self.setup_commands + \
-            self.commands + self.cleanup_commands
-        for command in joined_commands:
+        def _run_single(command):
             command_name = str(command)
-            if command_name in skip:
-                continue
+            if command_name in (skip or []):
+                return False
             logging.debug('execute %s', command_name)
             command.execute(ctx)
+            return True
+
+        cleanup_commands = []
+        try:
+            for cmd in self.setup_commands:
+                if _run_single(cmd[0]) and cmd[1]:
+                    cleanup_commands.insert(0, cmd[1])
+            for cmd in self.commands:
+                _run_single(cmd)
+        finally:
+            for cmd in cleanup_commands:
+                _run_single(cmd)
 
 
 class Command:
@@ -168,6 +179,7 @@ class SetupHome(Command):
         'AWS_SHARED_CREDENTIALS_FILE',
         'AWS_WEB_IDENTITY_TOKEN_FILE',
         'NETRC_FILE',
+        'REGISTRY_AUTH_FILE',
     ]
 
     def __init__(self):
@@ -180,33 +192,19 @@ class SetupHome(Command):
     def __str__(self):
         return 'setup_home'
 
-    @staticmethod
-    def _on_ci():
-        """
-            Detects if we are running on a CI system.
-            Returns the name of the CI system or None.
-        """
-        if os.environ.get('GITHUB_ACTIONS', False) == 'true':
-            return 'GitHub Actions'
-        elif os.environ.get('GITLAB_CI', False) == 'true':
-            return 'GitLab CI'
-        return None
-
     @staticmethod
     def _ssh_config_present():
         """
-            Checks if any file in the .ssh dir exists or
-            any manual ssh config option is set.
+            Checks if the .ssh/config file exists or any manual ssh config
+            option is set.
         """
         ssh_vars = ['SSH_PRIVATE_KEY', 'SSH_PRIVATE_KEY_FILE', 'SSH_AUTH_SOCK']
         if any(e in os.environ for e in ssh_vars):
             return True
 
         ssh_path = os.path.expanduser('~/.ssh')
-        if os.path.isdir(ssh_path):
-            with os.scandir(ssh_path) as it:
-                if any(it):
-                    return True
+        if os.path.exists(os.path.join(ssh_path, 'config')):
+            return True
         return False
 
     def _setup_netrc(self):
@@ -220,6 +218,30 @@ class SetupHome(Command):
                           'login gitlab-ci-token\n'
                           'password ' + os.environ['CI_JOB_TOKEN'] + '\n')
 
+    def _setup_registry_auth(self):
+        os.makedirs(self.tmpdirname + "/.docker")
+        if os.environ.get('REGISTRY_AUTH_FILE', False):
+            shutil.copy(os.environ['REGISTRY_AUTH_FILE'],
+                        self.tmpdirname + "/.docker/config.json")
+        elif not os.path.exists(self.tmpdirname + '/.docker/config.json'):
+            with open(self.tmpdirname + '/.docker/config.json', 'w') as fds:
+                fds.write("{}")
+
+        if os.environ.get('CI_REGISTRY', False) \
+                and os.environ.get('CI_JOB_TOKEN', False) \
+                and os.environ.get('CI_REGISTRY_USER', False):
+            with open(self.tmpdirname + '/.docker/config.json', 'r+') as fds:
+                data = json.loads(fds.read())
+                token = os.environ['CI_JOB_TOKEN']
+                base64_token = base64.b64encode(token.encode()).decode()
+                auths = data.get('auths', {})
+                auths.update(
+                    {os.environ['CI_REGISTRY']: {"auth": base64_token}})
+                data['auths'] = auths
+                fds.seek(0)
+                fds.write(json.dumps(data, indent=4))
+                fds.truncate()
+
     def _setup_aws_creds(self):
         aws_dir = self.tmpdirname + "/.aws"
         conf_file = aws_dir + "/config"
@@ -247,23 +269,40 @@ class SetupHome(Command):
             shutil.copy(os.environ['AWS_WEB_IDENTITY_TOKEN_FILE'],
                         webid_token_file)
 
+    @staticmethod
+    def _setup_gitlab_ci_ssh_rewrite(config):
+        ci_host = os.environ.get('CI_SERVER_HOST', None)
+        ci_port = os.environ.get('CI_SERVER_PORT', HTTPS_PORT_DEFAULT)
+        ci_prot = os.environ.get('CI_SERVER_PROTOCOL', 'https')
+        # added in GitLab 15.11. Set sensible defaults for older versions.
+        ci_ssh_host = os.environ.get('CI_SERVER_SHELL_SSH_HOST', ci_host)
+        ci_ssh_port = os.environ.get('CI_SERVER_SHELL_SSH_PORT',
+                                     SSH_PORT_DEFAULT)
+
+        for host in [ci_host, f'{ci_host}:{ci_port}']:
+            section = f'url "{ci_prot}://{host}/"'
+            config.add_value(section, 'insteadOf',
+                             f'git@{ci_ssh_host}:')
+            config.add_value(section, 'insteadOf',
+                             f'git@{ci_ssh_host}:{ci_ssh_port}')
+            config.add_value(section, 'insteadOf',
+                             f'ssh://git@{ci_ssh_host}/')
+            config.add_value(section, 'insteadOf',
+                             f'ssh://git@{ci_ssh_host}:{ci_ssh_port}/')
+
     def _setup_gitconfig(self):
         gitconfig_host = os.environ.get('GITCONFIG_FILE', False)
         gitconfig_kas = self.tmpdirname + '/.gitconfig'
 
-        # on supported CI systems, always try to read the gitconfig
-        if not gitconfig_host and self._on_ci():
+        # when running in a externally managed environment,
+        # always try to read the gitconfig
+        if not gitconfig_host and get_context().managed_env:
             gitconfig_host = os.path.expanduser('~/.gitconfig')
 
         if gitconfig_host and os.path.exists(gitconfig_host):
             shutil.copy(gitconfig_host, gitconfig_kas)
 
         with GitConfigParser(gitconfig_kas, read_only=False) as config:
-            # overwrite user as kas operates git
-            config['user'] = {
-                'email': 'kas@example.com',
-                'name': 'kas User'
-            }
             if os.environ.get('GIT_CREDENTIAL_HELPER', False):
                 config['credential'] = {
                     'helper': os.environ.get('GIT_CREDENTIAL_HELPER')
@@ -272,24 +311,21 @@ class SetupHome(Command):
                     config['credential']['useHttpPath'] = \
                         os.environ.get('GIT_CREDENTIAL_USEHTTPPATH')
             # in GitLab CI, add ssh -> https rewrites if no config is present
-            ci_server = os.environ.get('CI_SERVER_HOST', False)
-            if self._on_ci() == 'GitLab CI' and ci_server and \
+            ci_server = os.environ.get('CI_SERVER_HOST', None)
+            if get_context().managed_env == ME.GITLAB_CI and ci_server and \
                     not self._ssh_config_present() and \
                     not os.path.exists(gitconfig_host):
                 logging.debug('Adding GitLab CI ssh -> https rewrites')
-                section = f'url "https://{ci_server}/"'
-                config.add_value(section, 'insteadOf',
-                                 f'git@{ci_server}:')
-                config.add_value(section, 'insteadOf',
-                                 f'ssh://git@{ci_server}/')
+                self._setup_gitlab_ci_ssh_rewrite(config)
             config.write()
 
     def execute(self, ctx):
-        ci = self._on_ci()
-        if ci:
-            logging.info(f'Running on {ci}')
+        managed_env = get_context().managed_env
+        if managed_env:
+            logging.info(f'Running on {managed_env}')
         def_umask = os.umask(0o077)
         self._setup_netrc()
+        self._setup_registry_auth()
         self._setup_gitconfig()
         self._setup_aws_creds()
         os.umask(def_umask)
@@ -380,9 +416,10 @@ class WriteBBConfig(Command):
                 fds.write(ctx.config.get_bblayers_conf_header())
                 fds.write('BBLAYERS ?= " \\\n    ')
                 fds.write(' \\\n    '.join(
-                    sorted(_get_layer_path_under_topdir(ctx, layer)
-                           for repo in ctx.config.get_repos()
-                           for layer in repo.layers)))
+                          [_get_layer_path_under_topdir(ctx, layer)
+                           for repo in sorted(ctx.config.get_repos(),
+                                              key=lambda r: r.name)
+                           for layer in sorted(repo.layers)]))
                 fds.write('"\n')
                 fds.write('BBPATH ?= "${TOPDIR}"\n')
                 fds.write('BBFILES ??= ""\n')
@@ -408,9 +445,38 @@ class ReposApplyPatches(Command):
     def __str__(self):
         return 'repos_apply_patches'
 
+    def _vcs_operate_as_kas(self, gitconfig):
+        # currently only implemented for git
+        with GitConfigParser(gitconfig, read_only=False) as config:
+            # in case no user is defined, we keep the kas user
+            user_orig = {
+                'name': config.get('user', 'name',
+                                   fallback=KAS_USER_NAME),
+                'email': config.get('user', 'email',
+                                    fallback=KAS_USER_EMAIL)
+            }
+            config.set_value('user', 'name', KAS_USER_NAME)
+            config.set_value('user', 'email', KAS_USER_EMAIL)
+            config.write()
+            return user_orig
+
+    def _vcs_restore_user(self, gitconfig, user):
+        # currently only implemented for git
+        with GitConfigParser(gitconfig, read_only=False) as config:
+            config['user'] = user
+            config.write()
+
     def execute(self, ctx):
+        if 'HOME' not in ctx.environ:
+            raise ArgsCombinationError('Apply patches requires setup_home')
+
+        gitconfig = ctx.environ['HOME'] + '/.gitconfig'
+        user = self._vcs_operate_as_kas(gitconfig)
+
         repos_apply_patches(ctx.config.get_repos())
 
+        self._vcs_restore_user(gitconfig, user)
+
 
 class InitSetupRepos(Command):
     """
@@ -450,15 +516,16 @@ class SetupReposStep(Command):
             if repo_name not in ctx.config.get_repos_config():
                 # we don't have this repo yet (e.g. due to transitive incl.)
                 continue
-            ctx.missing_repos.append(ctx.config.get_repo(repo_name))
+            ctx.missing_repos.append((repo_name,
+                                      ctx.config.get_repo(repo_name)))
 
-        repos_fetch(ctx.missing_repos)
+        repos_fetch([v for k, v in ctx.missing_repos])
 
-        for repo in ctx.missing_repos:
+        for _, repo in ctx.missing_repos:
             repo.checkout()
 
         ctx.config.repo_dict.update(
-            {repo.name: repo for repo in ctx.missing_repos})
+            {id: repo for id, repo in ctx.missing_repos})
 
         repo_paths = {r: ctx.config.repo_dict[r].path for r
                       in ctx.config.repo_dict}
