@@ -33,7 +33,7 @@ import asyncio
 import errno
 import pathlib
 import signal
-from subprocess import Popen, PIPE
+from subprocess import Popen, PIPE, run as subprocess_run
 from .context import get_context
 from .kasusererror import KasUserError, CommandExecError
 
@@ -133,7 +133,8 @@ async def run_cmd_async(cmd, cwd, env=None, fail=True, liveupdate=False):
             cwd=cwd,
             env=env,
             stdout=asyncio.subprocess.PIPE,
-            stderr=asyncio.subprocess.PIPE)
+            stderr=asyncio.subprocess.PIPE,
+            preexec_fn=os.setpgrp)
     except FileNotFoundError as ex:
         if fail:
             raise ex
@@ -143,12 +144,26 @@ async def run_cmd_async(cmd, cwd, env=None, fail=True, liveupdate=False):
             raise ex
         return (errno.EPERM, str(ex))
 
+    # Process termination is a complicated thing. We need to ensure that
+    # the event-loop ThreadedChildWatcher thread fires before the loop is
+    # terminated. The best we can do it to ask the process to terminate
+    # (SIGINT) and wait for it. We need to shield the process wait to avoid
+    # that it is killed by the cancellation of the task, as we want a
+    # controlled termination. Forced terminations can leak an orphaned process.
+    # https://github.com/pytest-dev/pytest-asyncio/issues/708#issuecomment-1868488942
     tasks = [
         asyncio.ensure_future(_read_stream(process.stdout, logo.log_stdout)),
         asyncio.ensure_future(_read_stream(process.stderr, logo.log_stderr))
     ]
-    await asyncio.wait(tasks)
-    ret = await process.wait()
+
+    try:
+        await asyncio.gather(*[asyncio.shield(t) for t in tasks])
+        ret = await asyncio.shield(process.wait())
+    except asyncio.CancelledError:
+        process.terminate()
+        logging.debug('Command "%s" cancelled', cmdstr)
+        await process.wait()
+        raise
 
     if ret and fail:
         msg = f'Command "{cwd}$ {cmdstr}" failed'
@@ -162,17 +177,27 @@ async def run_cmd_async(cmd, cwd, env=None, fail=True, liveupdate=False):
     return (ret, ''.join(logo.stdout))
 
 
-def run_cmd(cmd, cwd, env=None, fail=True, liveupdate=False):
+def run_cmd(cmd, cwd, env=None, fail=True):
     """
         Runs a command synchronously.
     """
+    env = env or get_context().environ
+    cmdstr = ' '.join(cmd)
+    logging.debug('%s$ %s', cwd, cmdstr)
 
-    loop = asyncio.get_event_loop()
-    (ret, output) = loop.run_until_complete(
-        run_cmd_async(cmd, cwd, env, fail, liveupdate))
-    if ret and fail:
-        raise CommandExecError(cmd, ret)
-    return (ret, output)
+    try:
+        ret = subprocess_run(cmd, env=env, cwd=cwd, stdout=PIPE)
+        if ret.returncode and fail:
+            raise CommandExecError(cmd, ret.returncode)
+    except FileNotFoundError as ex:
+        if fail:
+            raise ex
+        return (errno.ENOENT, str(ex))
+    except PermissionError as ex:
+        if fail:
+            raise ex
+        return (errno.EPERM, str(ex))
+    return (ret.returncode, ret.stdout.decode('utf-8'))
 
 
 def find_program(paths, name):
@@ -189,6 +214,8 @@ def find_program(paths, name):
 def repos_fetch(repos):
     """
         Fetches the list of repositories to the kas_work_dir.
+
+        .. note:: termination point of the asyncio event loop.
     """
     if len(repos) == 0:
         return
@@ -207,6 +234,8 @@ def repos_fetch(repos):
 def repos_apply_patches(repos):
     """
         Applies the patches to the repositories.
+
+        .. note:: termination point of the asyncio event loop.
     """
     if len(repos) == 0:
         return
@@ -409,6 +438,7 @@ def ssh_no_host_key_check():
 def setup_parser_common_args(parser):
     from kas.libcmds import Macro
 
+    setup_cmds = [str(s) for (s, _) in Macro().setup_commands]
     parser.add_argument('config',
                         help='Config file(s), separated by colon. Using '
                         '.config.yaml in KAS_WORK_DIR if none is specified.',
@@ -419,7 +449,7 @@ def setup_parser_common_args(parser):
                         default=[],
                         action='append',
                         metavar='STEP',
-                        choices=[str(c) for c in Macro().setup_commands])
+                        choices=setup_cmds)
     parser.add_argument('--force-checkout', action='store_true',
                         help='Always checkout the desired commit/branch/tag '
                         'of each repository, discarding any local changes')
@@ -466,6 +496,7 @@ def setup_parser_keep_config_unchanged_arg(parser):
     steps = [
         'setup_dir',
         'finish_setup_repos',
+        'repos_checkout',
         'repos_apply_patches',
         'write_bbconfig',
     ]
