@@ -26,14 +26,21 @@
 import re
 import os
 import sys
+import linecache
 import logging
 import shutil
+from datetime import datetime
 from urllib.parse import urlparse
 from tempfile import TemporaryDirectory
 from .context import get_context
 from .libkas import run_cmd_async, run_cmd
 from .kasusererror import KasUserError
 
+if sys.version_info < (3, 8):
+    from cached_property import cached_property
+else:
+    from functools import cached_property
+
 __license__ = 'MIT'
 __copyright__ = 'Copyright (c) Siemens AG, 2017-2018'
 
@@ -119,9 +126,13 @@ class Repo:
                 continue
         return self.url
 
-    @property
+    @cached_property
     def revision(self):
         if self.commit:
+            (_, output) = run_cmd(self.get_commit_cmd(),
+                                  cwd=self.path, fail=False)
+            if output:
+                return output.strip()
             return self.commit
         if self.tag:
             (_, output) = run_cmd(self.resolve_tag_cmd(),
@@ -138,7 +149,7 @@ class Repo:
             return branch
         return None
 
-    @property
+    @cached_property
     def dirty(self):
         if not self.url:
             return True
@@ -240,6 +251,10 @@ class Repo:
             disable_operations = True
 
         if repo_type == 'git':
+            if commit and not re.match(r'^[0-9a-f]{40}|[0-9a-f]{64}$', commit):
+                logging.warning(
+                    f'{commit} is not a full-length hash for repo '
+                    f'"{name}". This will be an error in future versions.')
             return GitRepo(name, url, path, commit, tag, branch, refspec,
                            layers, patches, disable_operations)
         if repo_type == 'hg':
@@ -255,11 +270,14 @@ class Repo:
     def get_root_path(path, fallback=True):
         """
             Checks if path is under version control and returns its root path.
+            If the repo is a submodule, the root path of the super-repository
+            is returned.
         """
-        (ret, output) = run_cmd(['git', 'rev-parse', '--show-toplevel'],
-                                cwd=path, fail=False)
+        git_cmd = ['git', 'rev-parse', '--show-toplevel',
+                   '--show-superproject-working-tree']
+        (ret, output) = run_cmd(git_cmd, cwd=path, fail=False)
         if ret == 0:
-            return output.strip()
+            return sorted(output.strip().split('\n'))[0]
 
         (ret, output) = run_cmd(['hg', 'root'],
                                 cwd=path, fail=False)
@@ -337,7 +355,16 @@ class RepoImpl(Repo):
                              self.commit or self.tag or self.branch
                              or self.refspec,
                              output.strip())
-                return retc
+                # if branch is specified, check if it contains the commit
+                # also in our local clone
+                depth = get_context().repo_clone_depth
+                if self.branch and self.commit and not depth:
+                    (_, output) = await run_cmd_async(
+                        self.branch_contains_ref(), cwd=self.path, fail=False)
+                    if output.strip():
+                        return retc
+                else:
+                    return retc
 
         # Try to fetch if commit/tag/branch/refspec is missing or if --update
         # argument was passed
@@ -425,6 +452,10 @@ class RepoImpl(Repo):
         if self.operations_disabled or not self._patches:
             return 0
 
+        if self.dirty:
+            logging.warning(f'Repo {self.name} is dirty - no patching')
+            return 0
+
         (retc, _) = await run_cmd_async(self.prepare_patches_cmd(),
                                         cwd=self.path)
 
@@ -483,9 +514,17 @@ class RepoImpl(Repo):
                 raise PatchApplyError('Could not add patched files. repo: '
                                       f'{self.name}, vcs output: {output})')
 
-            cmd = self.commit_cmd()
+            timestamp = self.get_patch_timestamp(path)
+            if not timestamp:
+                dt = datetime.fromtimestamp(os.path.getmtime(path))
+                timestamp = dt.astimezone().strftime(
+                    "%a, %d %b %Y %H:%M:%S %z")
+
+            env = get_context().environ.copy()
+            cmd = self.commit_cmd(env, 'kas <kas@example.com>', 'msg',
+                                  timestamp)
             (retc, output) = await run_cmd_async(
-                cmd, cwd=self.path, fail=False)
+                cmd, cwd=self.path, env=env, fail=False)
             if retc:
                 raise PatchApplyError('Could not commit patch changes. repo: '
                                       f'{self.name}, vcs output: {output})')
@@ -546,9 +585,10 @@ class GitRepo(RepoImpl):
             cmd.extend([self.effective_url, self.path])
         return cmd
 
-    def commit_cmd(self):
-        return ['git', 'commit', '-a', '--author', 'kas <kas@example.com>',
-                '-m', 'msg']
+    def commit_cmd(self, env, author, msg, date):
+        env["GIT_COMMITTER_DATE"] = date
+        return ['git', 'commit', '-a', '--author', author, '-m', msg,
+                '--date', date]
 
     def contains_refspec_cmd(self):
         branch = self.branch or self.refspec
@@ -618,7 +658,14 @@ class GitRepo(RepoImpl):
         return ['git', 'remote', 'get-url', 'origin']
 
     def get_commit_cmd(self):
-        return ['git', 'rev-parse', '--verify', 'HEAD']
+        rev = self.commit or 'HEAD'
+        return ['git', 'rev-parse', '--verify', rev]
+
+    def get_patch_timestamp(self, path):
+        date = linecache.getline(path, 3)
+        linecache.clearcache()
+        if date and date.startswith("Date: "):
+            return date.replace("Date: ", "").strip()
 
 
 class MercurialRepo(RepoImpl):
@@ -639,8 +686,8 @@ class MercurialRepo(RepoImpl):
             return ['true']
         return ['hg', 'clone', self.effective_url, self.path]
 
-    def commit_cmd(self):
-        return ['hg', 'commit', '--user', 'kas <kas@example.com>', '-m', 'msg']
+    def commit_cmd(self, env, author, msg, date):
+        return ['hg', 'commit', '--user', author, '-m', msg, '--date', date]
 
     def contains_refspec_cmd(self):
         return ['hg', 'log', '-r', self.commit or self.tag or self.branch
@@ -673,7 +720,9 @@ class MercurialRepo(RepoImpl):
         return cmd
 
     def prepare_patches_cmd(self):
-        refspec = self.commit or self.tag or self.branch or self.refspec
+        refspec = (self.commit or self.tag or self.branch or self.refspec)
+        # strip revision part from refspec as not allowed in branch names
+        refspec = refspec.split(':')[-1]
         return ['hg', 'branch', '-f', f'patched-{refspec}']
 
     def apply_patches_file_cmd(self, path):
@@ -686,4 +735,14 @@ class MercurialRepo(RepoImpl):
         return ['hg', 'paths', 'default']
 
     def get_commit_cmd(self):
-        return ['hg', 'log', '-r', '.', '--template', '{node}\n']
+        rev = self.commit or '.'
+        return ['hg', 'log', '-r', rev, '--template', '{node}\n']
+
+    def get_patch_timestamp(self, path):
+        date = None
+        if linecache.getline(path, 3).startswith("# Date "):
+            date = linecache.getline(path, 4)
+        linecache.clearcache()
+
+        if date and date.startswith("# "):
+            return date.replace("# ", "").strip()
