@@ -29,10 +29,11 @@ import yaml
 import subprocess
 import pytest
 from kas import kas
-from kas.libkas import run_cmd
-from kas.libkas import TaskExecError, KasUserError
+from kas.libkas import TaskExecError, KasUserError, run_cmd
+from kas.attestation import file_digest_slow
 
 
+@pytest.mark.online
 def test_for_all_repos(monkeykas, tmpdir):
     tdir = str(tmpdir / 'test_commands')
     shutil.copytree('tests/test_commands', tdir)
@@ -49,6 +50,7 @@ def test_for_all_repos(monkeykas, tmpdir):
             == 'e9ca55a239caa1a2098e1d48773a29ea53c6cab2'
 
 
+@pytest.mark.online
 def test_for_all_repos_keep_config_unchanged(monkeykas, tmpdir):
     tdir = str(tmpdir / 'test_commands')
     shutil.copytree('tests/test_commands', tdir)
@@ -62,6 +64,7 @@ def test_for_all_repos_keep_config_unchanged(monkeykas, tmpdir):
     assert not os.path.exists("kas_1.1")
 
 
+@pytest.mark.online
 def test_checkout(monkeykas, tmpdir):
     tdir = str(tmpdir / 'test_commands')
     shutil.copytree('tests/test_commands', tdir)
@@ -86,6 +89,7 @@ def test_invalid_checkout(monkeykas, tmpdir, capsys):
         kas.kas(['checkout', 'test-invalid.yml'])
 
 
+@pytest.mark.online
 def test_checkout_with_ci_rewrite(monkeykas, tmpdir):
     tdir = str(tmpdir / 'test_commands')
     shutil.copytree('tests/test_commands', tdir)
@@ -97,17 +101,19 @@ def test_checkout_with_ci_rewrite(monkeykas, tmpdir):
         kas.kas(['checkout', 'test-url-rewrite.yml'])
 
 
+@pytest.mark.online
 def test_checkout_create_refs(monkeykas, tmpdir):
     tdir = str(tmpdir / 'test_commands')
     repo_cache = pathlib.Path(str(tmpdir.mkdir('repos')))
-    shutil.copytree('tests/test_patch', tdir)
+    shutil.copytree('tests/test_commands', tdir)
     monkeykas.chdir(tdir)
     monkeykas.setenv('KAS_REPO_REF_DIR', str(repo_cache))
     kas.kas(['checkout', 'test.yml'])
     assert os.path.exists(str(repo_cache / 'github.com.siemens.kas.git'))
-    assert os.path.exists('kas/.git/objects/info/alternates')
+    assert os.path.exists('kas_1.0/.git/objects/info/alternates')
 
 
+@pytest.mark.online
 def test_checkout_shallow(monkeykas, tmpdir):
     tdir = str(tmpdir / 'test_commands')
     shutil.copytree('tests/test_commands', tdir)
@@ -121,21 +127,21 @@ def test_checkout_shallow(monkeykas, tmpdir):
         mp.setenv('KAS_CLONE_DEPTH', '1')
         kas.kas(['checkout', 'test-shallow.yml'])
     for repo in ['kas_1', 'kas_2', 'kas_3', 'kas_4']:
-        (rc, output) = run_cmd(['git', 'rev-list', '--count', 'HEAD'],
-                               cwd=repo, fail=False, liveupdate=False)
-        assert rc == 0
+        output = subprocess.check_output(
+            ['git', 'rev-list', '--count', 'HEAD'], cwd=repo)
+        count = int(output.decode('utf-8').strip())
         if repo == 'kas_4':
-            assert output.strip() >= '1'
+            assert count >= 1
         else:
-            assert output.strip() == '1'
+            assert count == 1
 
 
+@pytest.mark.online
 def test_shallow_updates(monkeykas, tmpdir):
     def _get_commit(repo):
-        (rc, output) = run_cmd(['git', 'rev-parse', '--verify', 'HEAD'],
-                               cwd=repo, fail=False, liveupdate=False)
-        assert rc == 0
-        return output.strip()
+        output = subprocess.check_output(
+            ['git', 'rev-parse', '--verify', 'HEAD'], cwd=repo)
+        return output.decode('utf-8').strip()
 
     tdir = tmpdir / 'test_commands'
     tdir.mkdir()
@@ -170,6 +176,7 @@ def test_shallow_updates(monkeykas, tmpdir):
     assert _get_commit('kas') != commit
 
 
+@pytest.mark.online
 def test_repo_includes(monkeykas, tmpdir):
     tdir = str(tmpdir / 'test_commands')
     shutil.copytree('tests/test_repo_includes', tdir)
@@ -177,6 +184,7 @@ def test_repo_includes(monkeykas, tmpdir):
     kas.kas(['checkout', 'test.yml'])
 
 
+@pytest.mark.online
 def test_dump(monkeykas, tmpdir, capsys):
     tdir = str(tmpdir / 'test_commands')
     shutil.copytree('tests/test_repo_includes', tdir)
@@ -217,6 +225,7 @@ def test_dump(monkeykas, tmpdir, capsys):
                 assert os.path.exists('build/conf/local.conf')
 
 
+@pytest.mark.online
 def test_lockfile(monkeykas, tmpdir, capsys):
     tdir = str(tmpdir.mkdir('test_commands'))
     shutil.rmtree(tdir, ignore_errors=True)
@@ -235,6 +244,15 @@ def test_lockfile(monkeykas, tmpdir, capsys):
     kas.kas('dump --lock --inplace test.yml'.split())
     assert os.path.exists('test.lock.yml')
 
+    # check if legacy dump -> lock redirection works
+    with open('test.lock.yml', "rb") as f:
+        hash_dump = file_digest_slow(f, 'sha256')
+    os.remove('test.lock.yml')
+    kas.kas('lock test.yml'.split())
+    with open('test.lock.yml', "rb") as f:
+        hash_lock = file_digest_slow(f, 'sha256')
+    assert hash_dump.hexdigest() == hash_lock.hexdigest()
+
     # lockfile is considered during import, expect pinned branches
     kas.kas('dump test.yml'.split())
     lockspec = yaml.safe_load(capsys.readouterr().out)
@@ -340,3 +358,71 @@ def test_root_resolve_hg(monkeykas, tmpdir, capsys):
     kas.kas('dump --resolve-local test-local-hg.yml'.split())
     console = capsys.readouterr()
     assert 'contains uncommitted' not in console.err
+
+
+def test_ff_merges(monkeykas, tmpdir):
+    """
+    This tests check if kas correcly handles fast-forward merges.
+    """
+    tdir = tmpdir / 'test_commands'
+    rdirbare = tdir / 'upstream.git'
+    rdirwork = tdir / 'upstream.clone'
+    wdirkas = tdir / 'kas'
+    for d in [tdir, rdirbare, rdirwork, wdirkas]:
+        d.mkdir()
+
+    with open('tests/test_commands/test-ff-merges.yml', 'r') as f:
+        kas_input = yaml.safe_load(f)
+
+    shutil.copy('tests/test_commands/oe-init-build-env', rdirwork)
+    # create a bare upstream repository
+    monkeykas.chdir(rdirbare)
+    subprocess.check_call(['git', 'init', '--bare'])
+    subprocess.check_call(['git', 'symbolic-ref', 'HEAD', 'refs/heads/main'])
+    # create a repository with a branch (working copy)
+    monkeykas.chdir(rdirwork)
+    subprocess.check_call(['git', 'init'])
+    subprocess.check_call(['git', 'branch', '-m', 'main'])
+    subprocess.check_call(['git', 'add', 'oe-init-build-env'])
+    subprocess.check_call(['git', 'commit', '-m', 'test'])
+    # get head of main branch (before merge)
+    c_main = subprocess.check_output(['git', 'rev-parse', '--verify', 'HEAD'])
+    subprocess.check_call(['git', 'checkout', '-b', 'feature'])
+    subprocess.check_call(['touch', 'foo'])
+    subprocess.check_call(['git', 'add', 'foo'])
+    subprocess.check_call(['git', 'commit', '-m', 'test-feature'])
+    # get head of feature branch
+    c_feat = subprocess.check_output(['git', 'rev-parse', '--verify', 'HEAD'])
+    subprocess.check_call(['git', 'remote', 'add', 'origin', rdirbare])
+    subprocess.check_call(['git', 'push', 'origin', 'main', 'feature'])
+
+    # perform initial kas checkout
+    monkeykas.chdir(wdirkas)
+    kas_input['repos']['upstream']['commit'] = c_main.decode('utf-8').strip()
+    kas_input['repos']['upstream']['url'] = str(rdirbare)
+    with open('kas.yml', 'w') as f:
+        yaml.dump(kas_input, f)
+    kas.kas(['checkout', 'kas.yml'])
+
+    # ff merge feature into main
+    monkeykas.chdir(rdirwork)
+    # checkout main branch with kas
+    subprocess.check_call(['git', 'checkout', 'main'])
+    subprocess.check_call(['git', 'merge', '--ff-only', 'feature'])
+    subprocess.check_call(['git', 'push', 'origin', 'main'])
+
+    # bump commit on main branch in kas project, perform kas checkout
+    monkeykas.chdir(wdirkas)
+    # checkout main branch with kas again
+    kas_input['repos']['upstream']['commit'] = c_feat.decode('utf-8').strip()
+    with open('kas.yml', 'w') as f:
+        yaml.dump(kas_input, f)
+    kas.kas(['checkout', 'kas.yml'])
+
+
+def test_cmd_not_found(monkeykas, tmpdir):
+    cmd = ['/usr/bin/kas-not-exists']
+    ret, _ = run_cmd(cmd, tmpdir, os.environ, fail=False)
+    assert ret != 0
+    with pytest.raises(FileNotFoundError):
+        run_cmd(cmd, tmpdir, os.environ, fail=True)
